/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    SubscriptionService.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.IfmapConstStrings;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Link;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaState;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.exceptions.AlreadyObservedException;
import de.fhhannover.inform.iron.mapserver.exceptions.NoSuchSubscriptionException;
import de.fhhannover.inform.iron.mapserver.exceptions.PollResultsTooBigException;
import de.fhhannover.inform.iron.mapserver.exceptions.ResponseCreationException;
import de.fhhannover.inform.iron.mapserver.exceptions.SearchResultsTooBigException;
import de.fhhannover.inform.iron.mapserver.logging.Logging;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.messages.SearchRequest;
import de.fhhannover.inform.iron.mapserver.messages.SearchResult;
import de.fhhannover.inform.iron.mapserver.messages.SubPublishRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubSubscribeRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeDelete;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeUpdate;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;

/**
 * Class to handle SubscriptionRequests
 * 
 * @version 0.1
 * @author aw
 *
 */

/*
 * created: 30.04.10
 * 
 * changes:
 *  30.04.10 aw - created just a dummy class
 * 	28.05.10 vp - create startident if it's missing
 *	08.06.10 aw - some commenting and fixes
 *  12.06.10 aw - subscriptions with new notifyresult shall only notify the
 *   observer during updateSubscriptions
 *  30.09.10 aw - added rollbackChanges(), but looks like it is misplaced here.
 * 
 */
public class SubscriptionService {
	
	private static Logger logger;
	
	static {
		logger = Logging.getTheLogger();
	}
	
	private final PublisherRep publisherRep;
	private final MetaController metaController;
	private final IdentifierRep identifierRep;
	
	private SubscriptionObserver observer;
	
	private final Set<Subscription> changedSubscriptions;
	private final Set<Publisher> changedPublishers;
	private Map<Integer, Metadata> changedMetadata;

	SubscriptionService(IdentifierRep iRep, PublisherRep pRep, MetaController mCtl) {
		identifierRep	= iRep;
		publisherRep	= pRep;
		metaController	= mCtl;
		
		changedSubscriptions = new HashSet<Subscription>();
		changedPublishers = new HashSet<Publisher>();
		changedMetadata = new HashMap<Integer, Metadata>();
	}
	
	/**
	 * Handling of a {@link SubscriptionRequest} which is forwarded by the
	 * {@link DataModelService}.
	 * Look at all teh {@link SubPublishRequest} objects contained in the
	 * {@link SubscriptionRequest} and dispatch whether it's a update or
	 * a delete.
	 * 
	 * @param subscriptionReq
	 * @throws NoSuchSubscribeException
	 */
	void subscribe(SubscribeRequest subscribeReq) throws NoSuchSubscriptionException {
		
		Publisher publisher = publisherRep.getPublisherBySessionId(subscribeReq.getSessionId());
		
		for (SubSubscribeRequest ssr : subscribeReq.getSubSubscribeRequests()) {
			if (ssr instanceof SubscribeUpdate) {
				
					processSubscribeUpdate(publisher, (SubscribeUpdate)ssr);
					
			} else if (ssr instanceof SubscribeDelete) {
				
					processSubscribeDelete(publisher, (SubscribeDelete)ssr);
					
			}
		}
	}

	/**
	 * Delete a {@link Subscription} from a {@link Publisher}
	 * 
	 * @param publisher
	 * @param ssr
	 * @throws NoSuchSubscriptionException 
	 */
	private void processSubscribeDelete(Publisher publisher, SubscribeDelete ssr)
											throws NoSuchSubscriptionException {
		Map<String, Subscription> subscriptions = publisher.getSubscriptions();
		
		Subscription sub = subscriptions.get(ssr.getName());
		
		if (sub != null) {
			logger.debug("Deleting subscription \"" + sub.getName() + "\" of " + publisher);
			clearContainers(sub);
			subscriptions.remove(ssr.getName());
		} else {
			throw new NoSuchSubscriptionException("Subscription " +
					ssr.getName() + " does not exist");
		}
	}

	
	
	/**
	 * Create a new {@link Subscription}, attach it to a {@link Publisher},
	 * run the first search.
	 * 
	 * If a {@link Subscription} with the same name is already there, reset
	 * it and update the {@link SearchResult} objects
	 * 
	 * @param publisher
	 * @param ssr
	 */
	private void processSubscribeUpdate(Publisher publisher, SubscribeUpdate ssr) {
		
		logger.debug("Updating subscription \"" + ssr.getName() + "\" of "
				+ publisher.getPublisherId());
		
		if (publisher.pollResGrewTooBig()) {
			logger.warn("Ignoring subscribeUpdate as the PollResult grew too big"
					+ " and no notification was sent out Yestyet");
			return;
		}
		
		Subscription sub = publisher.getSubscriptions().get(ssr.getName());
	
		// subscription was already there
		if (sub != null) {
			logger.trace("Updating old subscription");
			if (sub.grewTooBig()) {
				logger.warn("Ignoring request, old subscription results grew to big!");
				return;
			}
			sub.setSearchRequest(ssr.getSearchRequest());
			sub.resetResults();
			clearContainers(sub);
		} else {
			sub = new Subscription(publisher, ssr.getName(), ssr.getSearchRequest());
			publisher.getSubscriptions().put(ssr.getName(), sub);
		}
		
		// if a startident is missing in the map, it must be created
		identifierRep.getOrCreateIdentifier(ssr.getSearchRequest().getStartIdentifier());

		// first get references to the real identfiers and links in the graph.
		SearchResultReferenceImpl sRes = new SearchResultReferenceImpl(publisher.getSessionId());
		SearchResultBuilder builder = new SearchResultBuilderImpl(sRes);
		Searcher searcher = new Searcher(identifierRep, ssr.getSearchRequest(), builder);
		
		try {
			searcher.runSearch();
			// connect the new subscription with them, so we can find it later
			// and construct the first search result
			builder = new SearchResultBuilderImpl(sub.getSearchResult());
			for (MetadataContainer mc : sRes.getContainers()) {
				// managing the list of subscriptions per Identifier and Link
				mc.addSubscription(sub);
				sub.addContainer(mc);
	
				// add the metadata into the search result
				builder.addMetadataContainer(mc);
				for (Metadata m : sRes.getMetadataFor(mc))
					builder.addMetadata(mc, m);
			}
		} catch (SearchResultsTooBigException e) {
			sub.setGrewTooBig();
		}

		sub.setChanged();
		addToChangedSubscriptions(sub);
		notifyChangedSubscriptions();
		cleanup();
	}
	

	private void cleanup() {
		changedSubscriptions.clear();
		changedPublishers.clear();
		changedMetadata.clear();
	}

	void updateSubscriptions(final Map<Integer, Metadata> changes) {
		logger.debug("Update Subscriptions...");
		changedMetadata = changes;
		logger.trace("Changed metadata:");
		for (Metadata m : changedMetadata.values()) {
			logger.trace("\t" + m.getPrefixAndElement() 
					+ " [state=" + m.getState()
					+ " cardinality=" + m.getCardinality()
					+ " lifetime=" + m.getLifeTime()
					+ " size=" + m.getByteCount() + " bytes]");
		}
		commitChanges();
		markSubscriptionsGrewTooBig();
		markPublisherPollResultTooBig();
		notifyChangedSubscriptions();
		cleanup();
	}
	

	/**
	 * Rollback changes done in the graph.
	 * Every {@link Metadata} object which has state NEW has to be removed
	 * from the graph. Every {@link Metadata} object which has state DELETED
	 * needs to be set to state UNCHANGED.
	 */
	void rollbackChanges() {
		logger.debug("Rollback changes made to metadata");
		
		for (Metadata m : changedMetadata.values()) {
			if (m.isDeleted()) {
				m.setState(MetaState.UNCHANGED);
			} else if (m.isNew()) {
				metaController.removeReferencesTo(m);
			} else if (m.isNotify()) {
				metaController.removeReferencesTo(m);
			} else {
				throw new RuntimeException("Encountered unchanged Metadata during rollback. " +
						"Programming Error?");
			}
		}
	}
	
	/**
	 * This is farked...
	 * 
	 * FIXME: That needs to be fixed, it's unreadable.. and a lot of
	 * code looks redundant...
	 * 
	 * Note:
	 * We go through the changes list twice. Once to find all the the metadata
	 * which is to be put into notifyResults. The second time for updateResult
	 * and deleteResult.
	 * TODO: for metadata to appear in a notifyResult the state of the
	 *       subscriptions before <b>all</b> changes because of publishUpdate
	 *       or publishDelete is taken. It's been like this some time before,
	 *       but one could think of the following scenario:
	 *       - A publishUpdate introduces a link somewhere, such that the
	 *         subscription would be active for more identifiers and links.
	 *       - A consecutive notify of metadata would appear in the notifyResult
	 *         because of the "new state" of the subscription.
	 *       As mentioned above, currently only the old state of subscriptions
	 *       is used.
	 *       
	 * @param changedMetadata
	 * @throws ResponseCreationException 
	 */
	private void commitChanges() {
		logger.debug("Commit Changes...");
		
		// take care of the notify metadata and remove it from the graph after
		// wards
		handleNotifyMetadata();
		
		// find all Subscriptions that need to be rerun and add them to
		// the changed subscriptions. We don't care about them, as they
		// are checked later on.
		List<Subscription> rerunSubs = findReRunSubscriptions();
	
		// run all searches with the old graph for the subscriptions that need
		// to be rerun
		for (Subscription sub : rerunSubs) {
			addToChangedSubscriptions(sub);
			logger.trace("Rerunning search for subscription \"" + sub.getName() + "\"");
			String sessionId = sub.getSearchRequest().getSessionId();
		
			// build up the old search result in the subscription
			sub.setReRunOldResult(new SearchResultReferenceImpl(sessionId));
			SearchResultBuilder builder = new SearchResultBuilderImpl(sub.getReRunOldResult());
			Searcher searcher = new Searcher(identifierRep, sub.getSearchRequest(), builder);
			searcher.runSearchIgnoreSize();
			
			clearContainers(sub);
		}

		// go through all changed (new or deleted) metadata
		for (Metadata m : changedMetadata.values()) {
			
			if (m.isNew() || m.isDeleted()) {
				MetadataContainer mc = m.getContainerReference();
				List<Subscription> subList = mc.getSubscriptions();
				if (mc instanceof Identifier) {
					for (Subscription sub : subList) {
						if (m.matchesFilter(sub.getSearchRequest().getResultFilter())) {
							if (m.isNew()) {
								sub.getUpdateResult().addMetadata(mc , m);
								sub.setChanged();
								addToChangedSubscriptions(sub);
							} else if (m.isDeleted()) {
								sub.getDeleteResult().addMetadata(mc, m);
								sub.setChanged();
								addToChangedSubscriptions(sub);
							}
						}
					}
				} else if (mc instanceof Link) {
					for (Subscription sub : subList) {
						if (m.matchesFilter(sub.getSearchRequest().getMatchLinksFilter())
								&& m.matchesFilter(sub.getSearchRequest().getResultFilter())) {
							if (m.isNew()) {
								sub.getUpdateResult().addMetadata(mc , m);
								sub.setChanged();
								addToChangedSubscriptions(sub);
							} else if (m.isDeleted()) {
								sub.getDeleteResult().addMetadata(mc, m);
								sub.setChanged();
								addToChangedSubscriptions(sub);
							}
						}
					}
				}
			}
			
			// new metadata is set to state UNCHANGED
			// deleted metadata is removed from the graph
			if (m.isNew()) {
				m.setState(MetaState.UNCHANGED);
			} else if (m.isDeleted()) {
				metaController.removeReferencesTo(m);
			} else if (m.isUnchanged()){
				throw new RuntimeException("Unchanged Metadata during commit");
			} else if (m.isNotify()) {
				// no problem they shouldn't be in the graph anymore
			} else {
				throw new RuntimeException("Metadata with unknown state during commit");
			}
		}
	
		// HOOK IN RERUN SUBSCRIPTIONS!!!
		// Run all searches with the new graph
		// Compare the old and new SearchResult
		for (Subscription sub : rerunSubs) {
			String sessionId = sub.getSearchRequest().getSessionId();
			sub.setReRunNewResult(new SearchResultReferenceImpl(sessionId));
			SearchResultBuilder builder = new SearchResultBuilderImpl(sub.getReRunNewResult());
			Searcher searcher = new Searcher(identifierRep, sub.getSearchRequest(), builder);
			searcher.runSearchIgnoreSize();
			
			for (MetadataContainer mc : sub.getReRunNewResult().getContainers()) {
				mc.addSubscription(sub);
				sub.addContainer(mc);
			}
			
			compareAndGenerateNewResults(sub);
		}
	}

	private void addToChangedSubscriptions(Subscription sub) {
		logger.trace("Adding " + sub.getName() + " to changed subscriptions");
		NullCheck.check(sub, "sub is null");
		changedSubscriptions.add(sub);
		changedPublishers.add(sub.getPublisherReference());
	}

	/**
	 * Simply iterate over all changed metadata which is "only"
	 * notify {@link Metadata}.
	 */
	private void handleNotifyMetadata() {
		for (Metadata m : changedMetadata.values()) {
			if (m.isNotify()) {
				handleConcreteNotifyMetadata(m);
			}
		}
	}
	
	/**
	 * Get the {@link MetadataContainer} for a notify {@link Metadata} object
	 * and check if one of the subscriptions attached to the {@link MetadataContainer}
	 * matches. We have to distinguish between {@link Identifier} and {@link Link}
	 * because different filter combinations have to be used.
	 * 
	 * @param m
	 */
	private void handleConcreteNotifyMetadata(Metadata m) {
		MetadataContainer mc = m.getContainerReference();
		for (Subscription sub : mc.getSubscriptions()) {
			if (mc instanceof Identifier) {
				if (m.matchesFilter(sub.getSearchRequest().getResultFilter())) {
					logger.trace("Adding notify metadata " + m.getPrefixAndElement()
							+ " to subscription " + sub.getName());
					sub.getNotifyResult().addMetadata(mc, m);
					sub.setChanged();
					addToChangedSubscriptions(sub);
				}
			} else if (mc instanceof Link) {
				if (m.matchesFilter(sub.getSearchRequest().getMatchLinksFilter())) {
					if (m.matchesFilter(sub.getSearchRequest().getResultFilter())) {
						logger.trace("Adding notify metadata " + m.getPrefixAndElement()
								+ " to subscription " + sub.getName());
						sub.getNotifyResult().addMetadata(mc, m);
						sub.setChanged();
						addToChangedSubscriptions(sub);
					}
				}
			}
		}
		// notify metadata can simply be removed and also _has to be_
		metaController.removeReferencesTo(m);
	}

	/**
	 * Helper method which compares the Old and New SearchResult and appends
	 * the differences into the Delete and Update Result.
	 * 
	 * {@link Subscription} objects contain new and old {@link SearchResult}
	 * objects which are set if the {@link Subscription} is rerun.
	 * Those are of the type {@link SearchResultImplReferences}, so we can
	 * compare with ==, as we compare references.
	 * 
	 * @param sub
	 */
	private void compareAndGenerateNewResults(Subscription sub) {
		SearchResult newRes = sub.getReRunNewResult();
		SearchResult oldRes = sub.getReRunOldResult();
		
		NullCheck.check(newRes, "PROGRAMMING ERROR!");
		NullCheck.check(oldRes, "PROGRAMMING ERROR!");
		
		List<MetadataContainer> oldContainers = oldRes.getContainers();
		List<MetadataContainer> newContainers = newRes.getContainers();
		
		for (MetadataContainer oldContainer : oldContainers) {
			if (newContainers.contains(oldContainer)) {
				List<Metadata> oldMeta = oldRes.getMetadataFor(oldContainer);
				List<Metadata> newMeta = newRes.getMetadataFor(oldContainer);	// old == new
				
				// compare metadata
				for (Metadata m : oldMeta) {
					if (newMeta.contains(m)) {
						newMeta.remove(m);
					} else {
						// newMeta does not contain oldMetadata object -> deleteResult
						sub.getDeleteResult().addMetadata(oldContainer, m);
					}
				}
				// only those that are not in oldMeta are left in newMeta,
				// we add them to the update result of the subscription.
				for (Metadata m : newMeta) {
					sub.getUpdateResult().addMetadata(oldContainer, m);
				}
				
				// remove the oldContainer from newContainers
				// newContainers will contain only those containers that are
				// new?
				newContainers.remove(oldContainer);
			} else {
				for (Metadata m : oldRes.getMetadataFor(oldContainer)) {
					sub.getDeleteResult().addMetadata(oldContainer, m);
				}
			}
		}
	
		// only those that are not in oldContainers, but in newContainers
		for (MetadataContainer newContainer : newContainers) {
			for (Metadata m : newRes.getMetadataFor(newContainer)) {
				sub.getUpdateResult().addMetadata(newContainer, m);
			}
		}
	}

	
	
	
	/**
	 * Go through the list of {@link Metadata} objects that are NEW or DELETED,
	 * if such a objects is attached to a {@link Link} some checks have to be
	 * done to find out if the {@link Subscription} needs to be rerun.
	 * 
	 * If the {@link Metadata} object is new, and the {@link Identifier} objects
	 * of the link contain {@link Subscription} objects that the {@link Link} does
	 * not contain, and further the newly created {@link Metadata} object
	 * matches the matchlinks filter of such a {@link Subscription} the 
	 * {@link Subscription} needs to be rerun.
	 * 
	 * If a {@link Metadata} object was deleted, we need to check for every
	 * {@link Subscription} object of this {@link Link} whether another
	 * {@link Metadata} object of that {@link Link} matches the matchlinks filter
	 * of the current {@link Subscription}. If this is not the case the
	 * {@link Subscription} needs to be rerun.
	 * 
	 * @param changedMetadata the {@link Metadata} objects that have been marked as changed
	 * @return
	 */
	private List<Subscription> findReRunSubscriptions() {
		
		HashSet<Subscription> ret = new HashSet<Subscription>();
		
		for (Metadata m : changedMetadata.values()) {
			if (m.getContainerReference() instanceof Link) {
				Link link = (Link) m.getContainerReference();
				if (m.isNew()) {
					HashSet<Subscription> set = new HashSet<Subscription>(link.getIdent1().getSubscriptions());
					set.addAll(link.getIdent2().getSubscriptions());
					
					for (Subscription sub : link.getSubscriptions()) {
						if (set.contains(sub)) {
							set.remove(sub);
						}
					}
				
					for (Subscription sub : set) {
						if (m.matchesFilter(sub.getSearchRequest().getMatchLinksFilter())) {
							sub.setChanged();
							ret.add(sub);
						}
					}
				} else if (m.isDeleted()) {
					for (Subscription sub : link.getSubscriptions()) {
						List<Metadata> all = link.getMetadata();
						int count = 0;
						for (Metadata md : all) {
							if (!md.isDeleted() && md.matchesFilter(sub.getSearchRequest().getMatchLinksFilter())) {
								count++;
							}
						}
						
						if (count == 0) {
							ret.add(sub);
							sub.setChanged();
						}
					}
				}
			}
		}
		return new LinkedList<Subscription>(ret);
	}

	/**
	 * If a {@link SubscriptionObserver} calls getPollResultFor() the corresponding
	 * Publisher with the given sessionId is searched.
	 * 
	 * All of it's subscriptions are checked, to see if any of them has changes.
	 * If a subscription has changes either the search result, the update result,
	 * the delete result or the notify result has elements which might be of
	 * interest for the MAPC.
	 * 
	 * A {@link PollResult} is constructed which contains {@link SearchResult}
	 * objects for all the different result types and further for all
	 * subscriptions a client has.
	 * 
	 * thoughts:
	 * 	could this get too big?
	 * 
	 * @param sessionId
	 * @return
	 * @throws PollResultsTooBigException 
	 */
	public PollResult getPollResultFor(String sessionId) throws PollResultsTooBigException {
		boolean rmSubs = false;
		Publisher pub = publisherRep.getPublisherBySessionId(sessionId);
		Collection<Subscription> activeSubs = pub.getSubscriptions().values();
		
		List<SearchResult> slist = new LinkedList<SearchResult>();
		List<SearchResult> ulist = new LinkedList<SearchResult>();
		List<SearchResult> dlist = new LinkedList<SearchResult>();
		List<SearchResult> nlist = new LinkedList<SearchResult>();
	
		// a simple list of string for those subscriptions that exceeded the
		// limit.
		List<String> errlist = new LinkedList<String>();
	
		// simply send do a pollResultsTooBig
		if (pub.pollResGrewTooBig()) {
			pub.unsetPollResGrewTooBig();
			throw new PollResultsTooBigException("too big");
		}
	
		
		for (Subscription sub : activeSubs) {
			if (sub.hasChanged()) {
				if (sub.grewTooBig()) {
					logger.debug("ErrorResult for " + sub.getName());
					errlist.add(sub.getName());
					rmSubs = true;
					continue;
				}
				
				if (!sub.getSearchResult().isEmpty())
					slist.add(sub.getSearchResult());
				
				if (!sub.getUpdateResult().isEmpty()) 
					ulist.add(sub.getUpdateResult());
				
				if (!sub.getDeleteResult().isEmpty())
					dlist.add(sub.getDeleteResult());
				
				if (!sub.getNotifyResult().isEmpty())
					nlist.add(sub.getNotifyResult());
				
				sub.resetResults();
			}
		}
		
		if (rmSubs) {
			removeSubscriptions(pub);
		}
		
		pub.setNotified(false);
		
		PollResult pr = new PollResult(slist, ulist, dlist, nlist, errlist);
		
		return pr;
	}
	

	/**
	 * Set the {@link SubscriptionObserver} to the given one.
	 * 
	 * @param subObs
	 * @throws AlreadyObservedException if another {@link SubscriptionObserver} was
	 *  registered before
	 */
	public void setSubscriptionObserver(SubscriptionObserver subObs) throws AlreadyObservedException {
		if (observer != null)
			throw new AlreadyObservedException("Only one SubscriptionObserver is allowed!");
		
		observer = subObs;
	}

	
	
	
	
	
	
	/**
	 * Simple helper method which calls pollResultAvailable() on the observer
	 * with the sessionId of the given publisher.
	 * 
	 * @param pub
	 */
	private void notifyChangedSubscriptions(Publisher pub) {
		if (observer != null) {
			
			// sanity check
			if (pub.getSessionId() == null) {
				throw new RuntimeException("PROGRAMMING ERROR: "
						+ "Publisher " + pub.getPublisherId() + " has no session");
			}
			
			if (!pub.wasNotified()) {
				logger.trace(pub.getPublisherId() + " has new poll results");
				pub.setNotified(true);
				observer.pollResultAvailable(pub.getSessionId());
			} else {
				logger.trace(pub + " was already notified!");
			}
		} else {
			throw new RuntimeException("SubscriptionObserver not set in Datamodel");
		}
	}
	
	
	/**
	 * Helper method to prepare the notification of the observer with all
	 * publishers that have new results available.
	 * 
	 * @param changedSubs
	 */
	private void notifyChangedSubscriptions() {
		logger.debug("Notify SubscriptionObserver about new results...");
		for (Publisher p : changedPublishers) {
			logger.trace(p.getPublisherId() + " has changed subscriptions");
			notifyChangedSubscriptions(p);
		}
	}

	
	/**
	 * Go through all changed subscriptions and check whether one grew too big.
	 * If one did, mark it as such and remove the subscription from the graph.
	 * This way it won't be checked the next time, but will still exist in the
	 * publisher.
	 */
	private void markSubscriptionsGrewTooBig() {
		for (Subscription sub : changedSubscriptions) {
			int sum = IfmapConstStrings.PRES_MIN_CNT;
			int maxSize = DataModelService.getServerConfiguration().getDefaultMaxSearchResultSize();
			SearchRequest sreq = sub.getSearchRequest();
			
			if (sreq.maxSizeGiven())
				maxSize = sreq.getMaxResultSize();

			sum += getTotalSize(sub);
			
			if (sum > maxSize)
				sub.setGrewTooBig();
			
			if (sub.grewTooBig()) {
				logger.trace("Subscription " + sub.getName() + " of " +
						sub.getPublisherReference().getPublisherId() + " got too big");
				clearContainers(sub);
			}
		}
	}

	/**
	 * If a publisher stores more than its set max-poll-result-size,
	 * he gets marked by us.
	 */
	private void markPublisherPollResultTooBig() {
		for (Publisher pub : changedPublishers) {
			int maxSize = DataModelService.getServerConfiguration().getDefaultMaxPollResultSize();
			int sum = IfmapConstStrings.PRES_MIN_CNT;;
			
			if (pub.maxPollSizeGiven()) {
				maxSize = pub.getMaxPollSize();
			}
			
			for (Subscription sub : pub.getSubscriptions().values())
				sum += getTotalSize(sub);
			
			if (sum > maxSize) {
				pub.setPollResGrewTooBig();
				removeSubscriptions(pub);
				logger.trace(pub.getPublisherId() + " PollResult grew too big " 
						+ " (" + sum + " bytes)");
			}
		}
	}

	private int getTotalSize(Subscription sub) {
		NullCheck.check(sub, "sub is null");
		int sum = 0;
		SearchResult sr = sub.getSearchResult();
		
		if (!sr.isEmpty())
			sum += IfmapConstStrings.SRES_MIN_CNT + sr.getByteCount();

		sr = sub.getNotifyResult();
		if (!sr.isEmpty())
			sum += IfmapConstStrings.NRES_MIN_CNT + sr.getByteCount();
		
		sr = sub.getUpdateResult();
		if (!sr.isEmpty())
			sum += IfmapConstStrings.URES_MIN_CNT + sr.getByteCount();
		
		sr = sub.getDeleteResult();
		if (!sr.isEmpty())
			sum += IfmapConstStrings.DRES_MIN_CNT + sr.getByteCount();
		
		return sum;
	}
	
	
	/**
	 * Remove all active subscriptions of the publisher with the given
	 * session-id.
	 * 
	 * @param sessionId
	 */
	void removeSubscriptions(Publisher pub) {
		NullCheck.check(pub, "pub is null");
		
		for (Subscription sub : pub.getSubscriptions().values()) {
			logger.trace("Clearing containers of subscription \"" + sub.getName() + "\"");
			clearContainers(sub);
		}
		pub.unsetPollResGrewTooBig();
		pub.setNotified(false);
		pub.clearSubscrptions();
	}
	

	/**
	 * Go through all {@link MetadataContainer} objects that this subscription
	 * is attached to and remove the reference to the given {@link Subscription}
	 * object.
	 * 
	 * @param sub
	 */
	void clearContainers(Subscription sub) {
		for (MetadataContainer mc : sub.getContainers()) {
			logger.trace("Removing subscription \"" + sub.getName() + "\" from " + mc);
			mc.removeSubscription(sub);
		}
		sub.getContainers().clear();
	}
}
