/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph
 * File:    SearchAble.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph;

import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;

/**
 * Objects which can show up in a search- or poll result have an associated
 * byte count. The byte count represent the number of bytes needed to represent
 * an object in a search or poll result.
 * 
 * Subclasses for now are {@link MetadataContainer} and {@link Metadata}. This
 * leads to the situation that all objects stored in the graph are subclasses
 * of {@link SearchAble} and can be queried for their byte count.
 * 
 * Convention: Instead of overwriting the {@link #getByteCount()} method in
 *             each subclass. A call to {@link #setByteCount(int)} has to be made
 *             in order to set the right value.
 * 
 * @author aw
 */
public abstract class SearchAble {
	
	/**
	 * indicates whether the byte count was set for this object.
	 */
	private boolean mByteCountSet;
	
	/**
	 * the byte count for this object.
	 */
	private int mByteCount;
	
	/**
	 * construct uninitialized.
	 */
	public SearchAble() {
		mByteCountSet = false;
		mByteCount = -1;
	}
	
	/**
	 * construct initialized
	 * 
	 * @param bytes
	 */
	public SearchAble(int bytes) {
		setByteCount(bytes);
	}
	
	protected final void setByteCount(int bytes) {
		if (bytes <= 0)
			throw new RuntimeException("byte count has to be greater than 0");
		
		mByteCountSet = true;
		mByteCount = bytes;
	}
	
	public final int getByteCount() {
		if (!mByteCountSet)
			throw new RuntimeException("access to uninitialized byte count");
		return mByteCount;
	}
}
