/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph.meta
 * File:    MetaType.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph.meta;

import java.util.HashMap;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.exceptions.InvalidMetadataException;
import de.fhhannover.inform.iron.mapserver.exceptions.MetadataConstructionException;
import de.fhhannover.inform.iron.mapserver.logging.Logging;

/**
 * A simple class that represents different MetaDataTypes.
 * Use it when comparing if Metadata is of the same type.
 * 
 * Metadata should have a typestring of urn:type
 * where urn might be http://example.com and type the Metadatatype
 * 
 * The standardmetadatatypes use:
 * urn:trustedcomputinggroup.org:2010:IFMAP-METADATA:2 as urn
 * and role, event, mac-ip ... as types.
 * 
 * @author aw
 * @version 0.1
 */
/*
 * 
 * created: 05.12.09
 * changes:
 * 	05.12.09 aw - First version of this class.
 *  17.12.09 aw - Fixed typestring not set
 *  10.04.10 aw - Added some logging...
 *  13.06.10 aw - Visibility change
 *
 */
public class MetaType {
	
	/**
	 * a static logger
	 */
	private static Logger logger;

	/**
	 * hashmap for all typestrings
	 */
	private static HashMap<Integer, MetaType> types;
	
	/**
	 * static initialization code
	 */
	static {
		logger = Logging.getTheLogger();
		types = new HashMap<Integer, MetaType>();
		initializeStandardMetadataCardinalities();
	}

	private final String typestring;
	
	private final MetaCardinalityType cardinality;
	
	private MetaType(String typestring, MetaCardinalityType card) {
		this.typestring = new String(typestring);
		cardinality = card;
	}
	
	/**
	 * For the set of standard metadata, the cardinalities are known.
	 * Create types for these when the class is first created.
	 * This way standard metadata cannot be publish with a inconsisten
	 * ifmap-cardinality value.
	 */
	private static void initializeStandardMetadataCardinalities() {
		for (String[] desc : StandardMetadata.STANDARD_METADATA) {
			try {
				instanceOfTypeWith(desc[0] + ":" + desc[1], desc[2]);
			} catch (MetadataConstructionException e) {
				logger.error("UNEXPECTED: Could not create MetaType for" +
						" standard metadata: " + e.getMessage());
			}
		}
	}
	
	/**
	 * Get the type for a given typestring.
	 * If a type with the same typestring exists this will be returned,
	 * else a new one will be constructed and stored.
	 * 
	 * The cardinality of the type is checked against a previous stored one.
	 * If the cardinality cannot be found or is inconsistent with any prior
	 * published cardinality for the same type throw an {@link MetadataConstructionException}
	 * to notify the client about this.
	 * 
	 * @param typestring
	 * @return
	 * @throws InvalidMetadataException 
	 * @throws MetadataConstructionException 
	 */
	static MetaType instanceOfTypeWith(String typestring, String card) throws
												MetadataConstructionException {
		MetaType t = null;
		if (typestring != null && card != null) {
				MetaCardinalityType cardinality;
				// parse the cardinality
				if (card.equals(MetaCardinalityType.singleValue.toString())) {
					cardinality = MetaCardinalityType.singleValue;
				} else if (card.equals(MetaCardinalityType.multiValue.toString())) {
					cardinality = MetaCardinalityType.multiValue;
				} else {
					// if we could not parse it, log it  and throw an exception
					logger.warn("Creation of MetaType with cardinality=" +
							card + " failed");
					throw new MetadataConstructionException("Invalid ifmap-cardinality value");
				}
			t = types.get(typestring.hashCode());
			if (t == null) {
				logger.trace("New MetaType with typestring=" + typestring +
						" and cardinality=" + card);
				t = new MetaType(typestring, cardinality);
				types.put(typestring.hashCode(), t);
			} else {
				// check if the given cardinality is equal to a already known
				// cardinality
				if (t.getCardinality() == cardinality) {
					// that's good!
				} else {
					// log a message and throw an exception if cardinality is
					// inconsistent
					logger.warn("Creation of MetaType with inconsistent " +
							"cardinality. Known is cardinality=" + t.getCardinality() +
							" new cardinality=" + cardinality);
					throw new MetadataConstructionException(
							"ifmap-cardinality is inconsistent with the " +
							"ifmap-cardinality stored for this type of metadata");
				}
			}
		} else {
			// one of the strings was null, most likely the ifmap-cardinality,
			// log it and throw an exception up.
			logger.warn("ifmap-cardinality was not given for this type");
			throw new MetadataConstructionException("No ifmap-cardinality?");
		}
		return t;
	}

	


	public String getTypeString() {
		return new String(typestring);
	}
	
	public MetaCardinalityType getCardinality() {
		return cardinality;
	}

	
	public boolean equals(Object o) {
		MetaType ot;
		if (o == null)
			return false;
		
		if (this == o)
			return true;
		
		if (o instanceof MetaType) {
			ot = (MetaType) o;
			return ot.getTypeString().equals(this.getTypeString());
		} else {
			return false;
		}
	}
}
